import tensorflow as tf
'''点乘'''
def test_matmul(a = 2,b = 3,c = 4, transpose_b = False):
    tips = """
    tf.matmul or tf.linalg.matmul
    a = 2
    b = 3
    c = 4
    transpose_b = False
    """
    
    x1 = tf.cast(tf.reshape(tf.range(a*b),(a,b)),dtype = tf.float32)
    
    x2 = tf.cast(tf.reshape(tf.range(a*b,b*(a+c)),(b,c)),dtype = tf.float32)
    
    if transpose_b:
        #x2 = tf.reshape(x2,(c,b))
        x2 = tf.transpose(x2)
        
    #ans = tf.matmul(x1,x2,transpose_b = transpose_b)
    ans = tf.linalg.matmul(x1,x2,transpose_b = transpose_b)
    print("ans:\n",ans)
if __name__ == "__main__":
    test_matmul(transpose_b = True)




'''
#查看gpu能否使用
旧版
print(tf.test.is_gpu_available())
新版
#可不加experimental
gpus = tf.config.experimental.list_physical_devices('GPU')
for gpu in gpus:
    tf.config.experimental.set_memory_growth(gpu, True)
'''

'''
import torch
#查看torch的gpu版本是否安装成功
print(torch.cuda.is_available())
'''

'''
#ubuntu命令行查看gpu使用情况
#nvidia-smi
#删除占用gpu内存进程
#kill -9 pid
'''

'''
#使用方法
旧版:
import tensorflow as tf
if tf.config.experimental.list_physical_devices('GPU'):
    from toolsbyerazhan.gputools import set_gpu_memory_tf
    set_gpu_memory_tf()

新版:
import toolsbyerazhan as tbe
tbe.set_gpu_memory_tf()
'''
def set_gpu_memory_tf():
    #一般要用到本函数的,都已经安装了tensorflow
    
    import tensorflow as tf
    if tf.config.experimental.list_physical_devices('GPU'):
        print("set gpu memory for tensorflow")
        gpus = tf.config.experimental.list_physical_devices('GPU')
        for gpu in gpus:
            tf.config.experimental.set_memory_growth(gpu, True)




import os

#后续将本地笔记中的os.py总结完成
#以及调整顺序，按照名称吧

def test_os__path__dirname():
    #得到当前文件的绝对路径
    print(os.path.dirname(os.path.abspath(__file__)))

def test_os__listdir(dir_path):
    #得到指定文件夹下所有的文件和文件夹
    for one in os.listdir(dir_path):
        print(dir_path + one)

def test_os__path__isfile_or_isdir(file_or_dir_path):
    #判断路径是文件还是文件夹
    if os.path.isfile(file_or_dir_path):
        print(file_or_dir_path,"is a file:")
    if os.path.isdir(file_or_dir_path):
        print(file_or_dir_path,"is a dir")

def test_os__path__exist(file_or_dir_path = "./ostools.py"):
    ans = os.path.exists(file_or_dir_path)
    print("file_or_dir_path",file_or_dir_path,"\nans:",ans)

if __name__ == "__main__":
    test_os__path__exist()



import pandas as pd
import os

def test_read_csv(file_name = "ipartment_cast_data.csv"):
    tips = """
tips:
    execute code:
    pd_data = pd.read_csv(file_name ,sep = ',',encoding = 'gbk')
    for i in range(len(pd_data)):
        print(i,list(pd_data.iloc[i]))
    """
    print(tips)
    file_path = os.path.dirname(os.path.abspath(__file__))
    file_name = os.path.join(file_path,file_name)

    print("最新的file_name",file_name)
    if not os.path.exists(file_name):
        test_to_csv(file_name)
        
    pd_data = pd.read_csv(file_name ,sep = ',',encoding = 'gbk')
    print("list(pd_data.columns):\n",list(pd_data.columns))
    print("\nlist(pd_data.iloc[i]):\n")
    for i in range(len(pd_data)):
        print(i,list(pd_data.iloc[i]))

def test_to_csv(file_name = None):
    tips = """
tips:
    execute code:
    pd_data.to_csv(file_name, sep = ',',index = False,encoding = 'gbk')
    the following parameters must be given:
    file_name,sep,index,encoding
    """
    if file_name is None:
        print(tips)
        file_name = "ipartment_cast_data.csv"
    
    columns = ["name","sex","birth year","howntown","occupation"]
    cast_data = [["陆展博","男",1986,"上海市","程序员"],
                 ["林宛瑜","女",1987,"淄博市","服装设计师"],
                 ["胡一菲","女",1988,"大连市","大学教授"],
                 ["曾小贤","男",1985,"福州市","电台主持人"],
                 ["吕子乔","男",1982,"成都市","自由职业者"],
                 ["陈美嘉","女",1985,"济南市","经纪人"],
                 ["关谷神奇","男",1985,"上海市","漫画家"],
                 ["张伟","男",1987,"鞍山市","律师"]
                ]
    pd_data = pd.DataFrame(cast_data,columns = columns)
    pd_data.to_csv(file_name, sep = ',',index = False,encoding = 'gbk')

if __name__ == "__main__":
    #test_df__to_csv()
    test_read_csv()



import tensorflow as tf
import numpy as np

class XLNetConfig(object):
    #model_type = "xlnet"
    #erazhan:最好还是从config.json中得到
    def __init__(
        self,
        vocab_size = 32000,
        d_model = 1024,
        n_layer = 24,
        n_head = 16,
        d_inner = 4096,
        ff_activation = "gelu",
        untie_r = True,#暂不
        attn_type = "bi",
        initializer_range = 0.02,#暂不
        layer_norm_eps = 1e-12,
        dropout = 0.1,
        mem_len = None,
        reuse_len = None,#暂不
        bi_data = False,#相对位置向量中
        clamp_len = -1,#同上
        same_length = False,
        summary_type = "last",#暂不
        summary_use_proj = True,#暂不
        summary_activation = "tanh",#暂不
        summary_last_dropout = 0.1,#暂不
        start_n_top = 5,#暂不
        end_n_top = 5,#暂不
        pad_token_id = 5,
        bos_token_id = 1,
        eos_token_id = 2,
        **kwargs
    ):
        
        self.vocab_size = vocab_size
        self.d_model = d_model
        self.n_layer = n_layer
        self.n_head = n_head
        assert d_model % n_head == 0,"d_model must be an integral multiple of n_head "
        if "d_head" in kwargs:
            assert (kwargs["d_head"] == d_model // n_head), "`d_head` ({kwargs['d_head']}) should be equal to `d_model // n_head` ({d_model // n_head})"
        self.d_head = d_model // n_head
        self.ff_activation = ff_activation
        self.d_inner = d_inner
        self.untie_r = untie_r
        self.attn_type = attn_type

        self.initializer_range = initializer_range
        self.layer_norm_eps = layer_norm_eps

        self.dropout = dropout
        self.mem_len = mem_len
        self.reuse_len = reuse_len
        self.bi_data = bi_data
        self.clamp_len = clamp_len
        self.same_length = same_length

        self.summary_type = summary_type
        self.summary_use_proj = summary_use_proj
        self.summary_activation = summary_activation
        self.summary_last_dropout = summary_last_dropout
        self.start_n_top = start_n_top
        self.end_n_top = end_n_top

        self.bos_token_id = bos_token_id
        self.pad_token_id = pad_token_id
        self.eos_token_id = eos_token_id

        if mem_len is None or mem_len == 0:
            #源码中用了warning.warn()
            print("Warning: This config doesn't use attention memories, a core feature of XLNet.")
            



import tensorflow as tf
import numpy as np

from tensorflow.keras.layers import Layer,Embedding
'''
说明：以下参考https://github.com/CyberZHG

keras-embed-sim：EmbeddingRet、EmbeddingSim
CyberZHG/keras-trans-mask：CreateMask
MaskEmbedding参考CyberZHG/keras-xlnet本身当中的mask_embed
keras-bert：Extract（很简单的操作）
keras_transformer_xl：Memory
'''

#其实如果用不到的话,不用构造这么麻烦,构造个基本能实现功能的就行了
class EmbeddingRet(Embedding):
    #除了Embedding正常的功能外,还会返回权重(self.embedding)
    #当前对应的shape也要修改
    def compute_output_shape(self,input_shape):
        return [super(Embedding,self).compute_output_shape(input_shape),
                (self.input_dim,self.output_dim),]

    def compute_mask(self,input,mask = None):
        '''
        Embedding中按照mask_zero是否为True来决定是否计算mask
        如果mask_zero=True就是需要让pad用0表示，计算mask时将
        非pad位标记为1，pad位标记为0
        '''
        return [super(EmbeddingRet,self).compute_mask(inputs,mask),None,]
    
    def call(self,inputs):
        return [super(EmbeddingRet,self).call(inputs),
                self.embeddings + 0.0,]
    #self.embedding是Embedding的属性

class EmbeddingSim(tf.keras.layers.Layer):
    """Calculate similarity between features and token embeddings with bias term."""
    #暂时还不知道在哪里用上
    def __init__(self,
                 use_bias=True,
                 initializer='zeros',
                 regularizer=None,
                 constraint=None,
                 stop_gradient=False,
                 **kwargs):
        """Initialize the layer.
        :param output_dim: Same as embedding output dimension.
        :param use_bias: Whether to use bias term.
        :param initializer: Initializer for bias.
        :param regularizer: Regularizer for bias.
        :param constraint: Constraint for bias.
        :param stop_gradient: Whether to stop gradient for input embedding.
        :param kwargs: Arguments for parent class.
        """
        super(EmbeddingSim, self).__init__(**kwargs)
        self.supports_masking = True
        self.use_bias = use_bias
        self.initializer = keras.initializers.get(initializer)
        self.regularizer = keras.regularizers.get(regularizer)
        self.constraint = keras.constraints.get(constraint)
        self.stop_gradient = stop_gradient
        self.bias = None

    def get_config(self):
        config = {
            'use_bias': self.use_bias,
            'initializer': keras.initializers.serialize(self.initializer),
            'regularizer': keras.regularizers.serialize(self.regularizer),
            'constraint': keras.constraints.serialize(self.constraint),
            'stop_gradient': self.stop_gradient,
        }
        base_config = super(EmbeddingSim, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def build(self, input_shape):
        if self.use_bias:
            embed_shape = input_shape[1]
            token_num = int(embed_shape[0])
            self.bias = self.add_weight(
                shape=(token_num,),
                initializer=self.initializer,
                regularizer=self.regularizer,
                constraint=self.constraint,
                name='bias',
            )
        super(EmbeddingSim, self).build(input_shape)

    def compute_output_shape(self, input_shape):
        feature_shape, embed_shape = input_shape
        token_num = embed_shape[0]
        return feature_shape[:-1] + (token_num,)

    def compute_mask(self, inputs, mask=None):
        if mask is None:
            return None
        return mask[0]

    def call(self, inputs, mask=None, **kwargs):
        inputs, embeddings = inputs
        if self.stop_gradient:
            embeddings = K.stop_gradient(embeddings)
        outputs = K.dot(inputs, K.transpose(embeddings))
        if self.use_bias:
            outputs = K.bias_add(outputs, self.bias)
        return keras.activations.softmax(outputs)

class CreateMask(Layer):
    """Create mask from input tensor.
    The shape of the mask equals to the shape of the input tensor.
    # Input shape
        Tensor with shape: `(batch_size, ...)`.
    # Output shape
        Tensor with shape: `(batch_size, ...)`.
    """

    def __init__(self, mask_value=0., **kwargs):
        super(CreateMask, self).__init__(**kwargs)
        self.supports_masking = True
        self.mask_value = mask_value

    def compute_output_shape(self, input_shape):
        return input_shape

    def compute_mask(self, inputs, mask=None):
        '''
        #这是官方Embedding中的compute_mask函数
        def compute_mask(self, inputs, mask=None):
            if not self.mask_zero:
                return None

            return math_ops.not_equal(inputs, 0)
        '''

        #K改为了tf
        return tf.not_equal(inputs, self.mask_value)

    def call(self, inputs, **kwargs):
        #K改为了tf，这个全0矩阵好像只是为了保持形状？
        return tf.zeros_like(inputs)

    def get_config(self):
        config = {'mask_value': self.mask_value}
        base_config = super(CreateMask, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

class RestoreMask(Layer):
    """Restore mask from the second tensor.
    # Input shape
        Tensor with shape: `(batch_size, ...)`.
        Tensor with mask and shape: `(batch_size, ...)`.
    # Output shape
        Tensor with shape: `(batch_size, ...)`.
    """

    def __init__(self, **kwargs):
        super(RestoreMask, self).__init__(**kwargs)
        self.supports_masking = True

    def compute_output_shape(self, input_shape):
        return input_shape[0]
    
    def compute_mask(self, inputs, mask=None):
        '''
        inputs = [token_embed,masking]
        mask = [mask0,mask1]
        最后的输出mask[1]=mask1的解释就是
        我们的输入inputs列表有两个元素
        一个是token_embed一个通过Createmask层计算的全为0的masking(仅用于传递新mask)
        那么因为self.supports_masking=True(都是支持mask传递的)
        两个输入在上一层计算中对应都有通过compute_mask函数计算得到
        的mask,那么我们将第2个输入masking对应的mask1作为
        第1个输入token_embed的新mask，因为token_embed原来的
        mask1(由EmbeddingRet得到，本质还是用Embedding计算)在mask_index!=0时
        计算为None
        '''
        return mask[1]
    #使用中,inputs=[token_embed,masking]
    #第一个是Embedding层后的词向量[batch_size,T,embed_size]
    #第二个是masking,[batch_size,T]全为0
    def call(self, inputs, **kwargs):
        return inputs[0] + 0.0

#这个是从
class MaskEmbedding(Layer):
    """Embedding for query tokens.

    # Arguments
        units: int >= 0. Number of hidden units.

    # Input shape
        Token embeddings, 3D tensor with shape: `(batch_size, seq_len, units)`.
        Query input, 2D tensor with shape: `(batch_size, seq_len)`.

    # Output shape
        3D tensor with shape: `(batch_size, seq_len, units)`.

    # References
        - [XLNet: Generalized Autoregressive Pretraining for Language Understanding](https://arxiv.org/pdf/1906.08237)
    """

    def __init__(self,
                 units,
                 initializer='uniform',
                 regularizer=None,
                 constraint=None,
                 **kwargs):
        super(MaskEmbedding, self).__init__(**kwargs)
        self.supports_masking = True
        self.units = units
        self.initializer = initializers.get(initializer)
        self.regularizer = regularizers.get(regularizer)
        self.constraint = constraints.get(constraint)

        self.embeddings = None

    def build(self, input_shape):
        #这个就是论文中的g_i^{(0)} = w
        self.embeddings = self.add_weight(
            shape=(1, 1, self.units),
            initializer=self.initializer,
            regularizer=self.regularizer,
            constraint=self.constraint,
            name='embeddings',
        )
        super(MaskEmbedding, self).build(input_shape)

    def compute_output_shape(self, input_shape):
        return input_shape[0]

    def compute_mask(self, inputs, mask=None):
        output_mask = None
        if mask is not None:
            output_mask = mask[0]
        return output_mask

    def call(self, inputs, **kwargs):
        token_embed, query = inputs
        query = tf.expand_dims(tf.cast(query,dtype = tf.float32),axis = -1)
        #query = K.expand_dims(K.cast(query, dtype=K.floatx()), axis=-1)

        #不理解怎么操作的啊，实际上query仅仅在预训练阶段使用，下游任务中是不需要的
        #[batch_size,seq_length,1]*[1,1,units]+[batch_size,seq_length]*[batch_size,seq_length,units]
        #query位置上为1(有几个位置？应该只有一个吧)，用专用的embedding，其它的位置上仍然用原来的token_embed
        #或者说query的位置上就是
        #因为训练时为了方便，是将一句话切分成两段，第二段的token作为预测token
        #query中这些预测位的token就是1，非预测位对应0,形式为[0,0,0,0,1,1](没有计算排列)
        #初始时预测位的token用一个可训练的vector代替本身wordembedding,
        #非预测位的token就还沿用本身的wordembedding
        #当然还要看最后的permutationmask怎么做，最后再总结一下
        #暂停
        return query * self.embeddings + (1.0 - query) * token_embed

    def get_config(self):
        config = {
            'units': self.units,
            'initializer': initializers.serialize(self.initializer),
            'regularizer': regularizers.serialize(self.regularizer),
            'constraint': constraints.serialize(self.constraint),
        }
        base_config = super(MaskEmbedding, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

class Extract(Layer):
    """Extract from index.
    See: https://arxiv.org/pdf/1810.04805.pdf
    """

    def __init__(self, index, **kwargs):
        super(Extract, self).__init__(**kwargs)
        self.index = index
        self.supports_masking = True

    def get_config(self):
        config = {
            'index': self.index,
        }
        base_config = super(Extract, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        return input_shape[:1] + input_shape[2:]

    def compute_mask(self, inputs, mask=None):
        return None

    def call(self, x, mask=None):
        return x[:, self.index]


class Memory(Layer):
    """Positional embeddings.
    # Arguments
        batch_size: int > 0. Maximum batch size.
        memory_len: int > 0. Maximum memory length.
        target_len: int > 0. Maximum length of targets.
        output_dim: int > 0. Dimension of outputs.
    # Input shape
        3D tensor with shape: `(batch_size, sequence_length, output_dim)`.
        1D tensor with shape: `(batch_size,)` represents length of memory.
    # Output shape
        3D tensor with shape: `(batch_size, sequence_length + memory_length, output_dim)`.
    # References
        - [Transformer-XL](https://arxiv.org/pdf/1901.02860.pdf)
    """

    def __init__(self, batch_size, memory_len, target_len, output_dim, **kwargs):
        super(Memory, self).__init__(**kwargs)
        self.supports_masking = True
        self.stateful = True#支持在不同的batch中进行传递信息，在推理中也会运行

        self.batch_size = batch_size
        self.memory_len = memory_len
        self.target_len = target_len
        self.output_dim = output_dim

        self.memory = None

    def build(self, input_shape):
        self.memory = self.add_weight(
            shape=(self.batch_size, self.memory_len + self.target_len, self.output_dim),
            initializer='zeros',
            trainable=False,#不可训练，是根据上一个batch最后一层计算得到的?
            name='memory',
        )
        super(Memory, self).build(input_shape)

    def compute_output_shape(self, input_shape):
        return input_shape[0][0], None, self.output_dim

    def compute_mask(self, inputs, mask=None):
        if mask is None:
            return None
        return mask[0]

    def call(self, inputs, **kwargs):
        #注意这里要把K换成tf等，否则保错
        #inputs：[batch_size,seq_length,hidden_size]
        #memory_length：[batch_siz,1]
        inputs, memory_length = inputs
        memory_length = tf.cast(memory_length[0][0], 'int32')
        batch_size = tf.cast(tf.shape(inputs)[0], 'int32')
        seq_len = tf.cast(tf.shape(inputs)[1], 'int32')

        # Build new memory
        pad = tf.tile(inputs[0:1, ...], (self.batch_size - batch_size, 1, 1))
        padded = tf.concatenate([inputs, pad], axis=0)              # (self.batch_size, seq_len, output_dim)
        new_memory = tf.concatenate([self.memory, padded], axis=1)  # (self.batch_size, self.memory_len + seq_len, ...)
        new_memory = tf.slice(                                     # (self.batch_size, self.memory_len, output_dim)
            new_memory,
            (0, seq_len, 0),
            (self.batch_size, self.memory_len + self.target_len, self.output_dim),
        )
        self.add_update(K.update(self.memory, new_memory), inputs)

        # Build output
        old_memory = tf.slice(                                     # (batch_size, memory_length, output_dim)
            new_memory,
            (0, K.maximum(0, self.memory_len + self.target_len - seq_len - memory_length), 0),
            (batch_size, K.minimum(self.memory_len, memory_length), self.output_dim),
        )

        return old_memory

    def get_config(self):
        config = {
            'batch_size': self.batch_size,
            'memory_len': self.memory_len,
            'target_len': self.target_len,
            'output_dim': self.output_dim,
        }
        base_config = super(Memory, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



import tensorflow as tf
import numpy as np

#按照函数名进行排序
def test_Activation(activation = 'relu'):
    from tensorflow.keras.layers import Activation
    tips = """
tips:
    params:
    activation:'relu'(default),'gelu','square'
    """
    
    def my_square(x):
        return x**2
        
    def gelu(x):
    	return 0.5 * x * (1.0 + tf.math.erf(x / np.sqrt(2.0)))

    if activation == 'square':
        ac_fun = my_square#自定义的函数要写函数名,不能加引号
    elif activation == 'gelu':
        ac_fun = gelu
    else:
        ac_fun = 'relu'
    layer = Activation(ac_fun)

    inputs = tf.convert_to_tensor([0,-1,2,-3,4,-5,6,-7,8,-9],dtype = tf.float32)

    print(tips)

    print("inputs.numpy():\n",inputs.numpy())

    output = layer(inputs)
    print("output.numpy():\n",output.numpy())

def test_Dot(a1 = 2, a2 = 2):
    from tensorflow.keras.layers import Dot
    tips= """
tips:
    将两个tensor进行point-wise点乘
    
    Dot参数:
    axes：两个tensor各自相乘的维度，比如[2,2]，就表示将x和y的第2维对应相乘（维度从0开始）
    normalize：是否是将两个向量相乘结果标准化，即求余弦相似度

    注意使用方式，Dot内部参数调整位置，需要将x,y写为列表形式[x,y]进行带入Dot层进行call运算

    test_Dot参数:
    axes = [a1,a2],本示例中,a1可为1或2，a2固定为2，它们对应的维度必须相等
    """
    x = tf.convert_to_tensor([[[1,1,1],[2,2,2],[3,3,3]]],dtype=tf.float32)
    y = tf.convert_to_tensor([[[3,3,3],[4,4,4]]],dtype=tf.float32)
    m = tf.keras.layers.Dot(axes=[a1,a2],normalize=False)([x,y])
    print(tips)
    print("x:\n",x)
    print("y:\n",y)
    print("m:\n",m)
    #print(m.numpy())

def test_Dropout(print_remarks = False):
    from tensorflow.keras.layers import Layer,Dropout
    tips = """
tips:
    Dropout层的实现过程:
    1、对每个元素按照概率rate进行丢弃
    2、对剩下没被丢弃的元素乘上1/(1 - rate)
    test_Dropout参数:
    print_remarks：是否print出remarks(理解Dropout时的总结)
    """
    remarks = """
remarks:
    注意在第二种写法中的training=True,它是在Dropout层中的
    call方法的参数,默认为None,如果设定为True,则会在测试
    中展现效果,否则它会自动判断是否在训练中，如果在训练中
    则会实现dropout效果

    https://zhuanlan.zhihu.com/p/32846476
    该网址很好的解释了tf.layers.dropout的原理（另外还讲了BN，后续详细看）

    https://blog.csdn.net/nini_coded/article/details/79302800
    原理是将inputs中每个元素按概率rate随机丢弃，剩下的乘上1/(1-rate)进行rescale
    开始理解错了，以为是从所有元素中选rate*N个进行丢弃
    """

    print(tips)
    if print_remarks:
        print(remarks)
        
    class MyDropout(Layer):
        '''自定义Dropout层'''
        def __init__(self, rate = None, **kwargs):
            super(MyDropout, self).__init__(**kwargs)
            self.rate = rate
        '''不要用自己层，因为暂时做不到让它自己判断是否在训练中，这里只是为了理解dropout的原理实现'''
    
        def call(self, inputs, training = False):

            if training == False:
                return inputs
        
            if self.rate is not None:
            
                mask = tf.random.uniform(shape = tf.shape(inputs),minval = 0, maxval = 1,dtype = tf.float32)
                mask = tf.cast(mask > self.rate, dtype = tf.float32)/(1 - self.rate)
                inputs = inputs * mask

            return inputs

    rate = 0.5
    
    inputs = tf.reshape(tf.range(1,25),(2,3,4))
    inputs = tf.cast(inputs, dtype = tf.float32)

    print("inputs.numpy()\n",inputs.numpy(),'\n')

    #y=tf.nn.dropout(x,rate=0.31)
    ans1 = Dropout(rate = rate)(inputs, training = True)
    ans2 = MyDropout(rate = rate)(inputs, training = True)

    print("MyDropout Layer:\n",ans1.numpy())
    print("Dropout Layer:\n",ans2.numpy())

    #print("tf.reduce_sum(inputs):",tf.reduce_sum(inputs).numpy())
    #print("tf.reduce_sum(ans1):",tf.reduce_sum(ans1).numpy())
    #print("tf.reduce_sum(ans2):",tf.reduce_sum(ans2).numpy())

def test_Reshape():
    from tensorflow.keras.layers import Reshape
    tips = """
tips:
    a.shape:[3,4,10]
    b = Reshape((-1,))(a)
    c = Reshape((5,8))(a)

remarks:
    Reshape层默认将第一维看作batch_size
    接口参数是不包括batch_size维的
    所以除了不考虑batch_size维度
    Reshape的规则和tf.reshape的规则是一样的
    """
    print(tips)
    a = tf.reshape(tf.range(120),(3,4,10))
    b = Reshape((-1,))(a)
    c = Reshape((5,8))(a)
    

    print("a.shape:",a.shape)
    print("b.shape:",b.shape)
    print("c.shape:",c.shape)

#按照函数名进行排序
if __name__ == "__main__":
    #import toolsbyerazhan as tbe
    #tbe.set_gpu_memory_tf()
    #test_Activation('square')
    #test_Dot(1,2)
    #test_Dropout(True)
    test_Reshape()
    pass



import json

def test_json__loads():
    
    json_str="""
{
    "gpsId" : "10014350",
    "businessId" : "12723026",
    "orderStatus" : { "$numberInt" : "10" },
    "sttDate" : { "$date" : { "$numberLong" : "1598274573000" } },
    "endDate" : { "$date" : { "$numberLong" : "1598572958000" } },
    "curDate" : { "$date" : { "$numberLong" : "1598572957000" } },
    "carNum" : "冀GC3543",
    "_class" : "com.techbrc.gps.trace.vo.DataInfo" }"""
    print("raw json_str",json_str)
    
    ans = json.loads(json_str)
    c = 0
    print("\nexecute code:\nans = json.loads(json_str)\nthe type of ans is dict")
    
    print("\nthe following is to travel ans\n")
    for k,v in ans.items():
        c += 1
        print(c,k,v)
if __name__ == "__main__":
    test_json__loads()



import tensorflow as tf
import numpy as np

from xlnet_layers import CreateMask
from utils import *

#[batch_size,seq_length]
inputs = tf.convert_to_tensor([[0,2,1],[0,1,1]],dtype = tf.int32)
masks = CreateMask(mask_value=1)(inputs)
print(inputs.numpy())
print(masks.numpy())



import tensorflow as tf

'''
原版本:tf.matrix_band_part
新版本:tf.linalg.band_part
作用是根据对角线的位置读取矩阵特定位置上的数据
可以作用在多维张量上，使用时参考官方文档的解释
二维的好理解一些

参数(二维):
input:就是矩阵
num_lower:整数,控制下三角,0就是只保留对角线元素不动,-1表示保留所有下三角元素,n=1表示保留对角线下n=1排
num_upper:整数,控制上三角,类似num_lower
name:操作名称

tf.linalg.band_part(inputs,0,0)就是只保留对角线的元素，其余元素均为0
tf.linalg.band_part(inputs,-1,-1)不做任何修改
'''

def test_band_part(lower = 0, upper = 0, r = 5, c = 6, use_ones = False):
    
    tips = '''
tip:
    params(default):
    lower:0
    upper:0
    r:5
    c:6
    use_ones:False

    examples:
    test_band_part()
    test_band_part(1,-1,3,5)
    test_band_part(1,2,use_ones = True)
    '''
    
    if use_ones:
        inputs = tf.ones([r,r],dtype = tf.float32)
    else:
        inputs = tf.reshape(tf.range(r*c),[r,c])
        inputs = tf.cast(inputs,dtype = tf.float32)
    print(tips)
    print("inputs.numpy():\n",inputs.numpy())
    
    ans = tf.linalg.band_part(inputs,lower,upper)
    print("ans.numpy():\n",ans.numpy())

if __name__ == "__main__":
    import toolsbyerazhan as tbe
    test_band_part()
    #pass



import tensorflow as tf
import numpy as np


def get_initializer(initializer_range=0.02):
    """Creates a `tf.initializers.truncated_normal` with the given range.
    Args:
        initializer_range: float, initializer range for stddev.
    Returns:
        TruncatedNormal initializer with stddev = `initializer_range`.
    """
    return tf.keras.initializers.TruncatedNormal(stddev=initializer_range)

def shape_list(x):
    """Deal with dynamic shape in tensorflow cleanly."""
    static = x.shape.as_list()
    dynamic = tf.shape(x)
    return [dynamic[i] if s is None else s for i, s in enumerate(static)]



import tensorflow as tf
import numpy as np

from tensorflow.keras.layers import Layer
#注:该部分代码从笔记的tensorflow文件夹中复制

'''
数据名称说明：
inputs:[batch_size,max_seq_length],tf.int32
targets:[batch_size,max_seq_length],tf.int32
input_length:[batch_size,],tf.int32
logits:[batch_size,max_seq_length,num_tags],tf.float32
transition_params:[num_tags,num_tags],tf.float32
'''

class CRFLayer(Layer):
    
    def __init__(self,num_tags,**kwargs):
        
        super(CRFLayer,self).__init__(**kwargs)
        self.num_tags = num_tags

    def build(self, input_shape):
        super(CRFLayer,self).build(input_shape)
        shape = (self.num_tags,self.num_tags)
        self.transition_params = self.add_weight(name = "transition_params",
                                                 shape = shape,
                                                 initializer = 'glorot_uniform',
                                                 trainable = True)
        
    def call(self, logits, transition_params = None):
        "input_length是要给定的,input代表最开始的输入,在bilstm的Embedding层接收的inputs" 
        
        self.batch_size, self.max_seq_length= tf.TensorShape(tf.shape(logits))[:2]
        assert self.num_tags == tf.TensorShape(tf.shape(logits))[-1], "inputs的维度与self.num_tags不匹配"
        
        #手动给定transition_params,用于测试CRFLayer效果
        if transition_params is not None:
            self.transition_params = transition_params
        
        return logits
    
    def crf_log_likelihood(self, logits, targets, input_lengths):

        unary_score = self.crf_unary_score(logits, targets, input_lengths)
        binary_score = self.crf_binary_score(targets, input_lengths)

        score = unary_score + binary_score
        log_norm = self.crf_log_norm(logits, input_lengths)
        loss = score - log_norm

        return loss

    def crf_log_norm(self, logits, input_lengths):

        transition_params = tf.expand_dims(self.transition_params, 0)
        
        _state = logits[:,0,:]

        last_index = tf.cast(tf.maximum(0, input_lengths - 1),dtype = tf.int32)
        all_alphas = tf.expand_dims(_state, 1)

        for i in range(1, self.max_seq_length):
            _input = logits[:,i,:]
            
            _state = tf.expand_dims(_state, 2)
            transition_scores = _state + transition_params

            _state = _input + tf.reduce_logsumexp(transition_scores, [1])

            all_alphas = tf.concat([all_alphas, tf.expand_dims(_state, 1)],axis = 1)

        idxs = tf.stack([tf.range(self.batch_size),last_index], axis = 1)

        #all_alphas维度为[batch_size,num_tags],每个batch的num_tags个元素代表的是最后一个时间步的分值
        all_alphas = tf.gather_nd(all_alphas, idxs)

        #log_norm的维度为[batch_size,]
        log_norm = tf.reduce_logsumexp(all_alphas, [1])

        #确保出现长度为0的输入对应的loss为0
        log_norm = tf.where(tf.less_equal(input_lengths, 0), tf.zeros_like(log_norm), log_norm)
        return log_norm
    
    def crf_unary_score(self, logits, targets, input_lengths):
        '''思路是将logits铺开，变成一个一维向量，维度维batch_size*max_seq_length*num_tags，
            然后构造根据targets计算出选择的位置'''
        '''求S1,输出维度维度为(batch_size,)'''
        batch_size,max_seq_length,num_tags = tf.TensorShape(tf.shape(logits))[:3]

        #维度由[batch_size,max_seq_length,num_tags]变成了batch_size*max_seq_length*num_tags
        #效果展示看test_func_1()
        flattened_logits = tf.reshape(logits, [-1])

        #维度最后是(batch_size,1)数字从0开始每次加max_seq_length*num_tags
        #效果展示看test_func_2()
        offsets = tf.expand_dims(tf.range(batch_size) * max_seq_length * num_tags, 1)

        #最后维度是(batch_size,max_seq_length),每个元素从0开始依次加num_tags
        #效果展示看test_func_3()
        offsets += tf.expand_dims(tf.range(max_seq_length) * num_tags,0)

        #这样做的目标完全就是为了带入tf.gather
        flattened_tag_offsets = tf.reshape(offsets + targets, [-1])

        unary_scores = tf.reshape(tf.gather(flattened_logits, flattened_tag_offsets), (batch_size,max_seq_length))

        #一定要转为tf.float32格式，否则最后计算时会报错
        #参考代码笔记中的tf.sequence_mask
        masks = tf.sequence_mask(input_lengths, maxlen = tf.shape(targets)[1], dtype = tf.float32)

        unary_scores = tf.reduce_sum(unary_scores * masks, 1)

        return unary_scores
    
    def crf_binary_score(self, targets, input_lengths):
        '''求S2，当然也可以用tf.einsum，但是个人偏好用这种更底层的方式'''

        #也可以用tf.slice切片
        start_targets = targets[:,:-1]
        end_targets = targets[:,1:]

        #建立相邻两个时间步状态对(yp,y)，计算平坦化后位置顺序，维度为[batch_size, max_seq_length - 1]
        start_to_end = start_targets * self.num_tags + end_targets

        #将转移概率矩阵平坦化
        flattened_transition_params = tf.reshape(self.transition_params, [-1])

        #根据start_to_end在flattened_transition_params中找相应位置的数字，维度和start_to_end相同
        binary_scores = tf.gather(flattened_transition_params, start_to_end)

        #根据长度构造masks，维度为[batch_size,max_seq_length]
        masks = tf.sequence_mask(input_lengths, maxlen = tf.shape(targets)[1],dtype = tf.float32)

        #将masks的维度匹配成和binary_scores一样，均为[batch_size,max_seq_length-1]
        #不用masks[:,:-1]的原因是start为True，但end可能为False,
        #start为True,end不一定为True，也可能 是False
        #end为True,start一定为True
        #end为False,那么就没必要计算这对评分值了
        
        masks = masks[:,1:]
        binary_scores = tf.reduce_sum(binary_scores * masks, 1)

        return binary_scores

    def viterbi_decode(self, logits):
        '''
        logits：[1,seq_length,num_tags]或者[seq_length,num_tags],如果是前者则将其转为后者,且用numpy进行操作
        return：path_scores，path_states
        '''
        n_dims = len(tf.shape(logits))

        assert n_dims == 2 or (n_dims == 3 and tf.shape(logits)[0] == 1), "n_dims必须为2或3,只处理一条数据"
        
        if n_dims == 3:
            
            logits = tf.squeeze(logits, 0)

        #score的维度为
        logits = np.array(logits, dtype = np.float32)
        
        transition_params = np.array(self.transition_params, dtype = np.float32)
        
        seq_length, num_tags = logits.shape
        
        pathScores = np.zeros_like(logits, dtype = np.float32)
        pathStates = np.zeros_like(logits, dtype = np.int32)
        
        pathScores[0] = logits[0]
    
        for i in range(seq_length - 1):

            scores = pathScores[i]
            temM = scores[:, np.newaxis] + transition_params
            
            pathScores[i + 1] = np.max(temM, axis = 0) + logits[i+1]
            pathStates[i + 1] = np.argmax(temM, axis = 0) 

        path = np.zeros(seq_length, dtype = np.int32)
        path[-1] = np.argmax(pathScores[-1])

        for i in range(seq_length - 1, 0, -1):
            path[i-1] = pathStates[i,path[i]]

        return list(path), np.max(pathScores[-1])
    
    def crf_decode(self):
        #处理一个batch的数据,后续补充,可采用RNNCell
        pass

if __name__=="__main__":
    pass



    



import tensorflow as tf
import numpy as np
from tensorflow.keras.layers import Layer,Embedding,Bidirectional,LSTM,Input,Dot,Reshape,Lambda,Dense,Dropout

#from tensorflow.keras import Sequential

class MVLSTM(tf.keras.Model):
    def __init__(self, hyper_params, *args, **kwargs):
        super(MVLSTM,self).__init__(*args,**kwargs)
        self.hp = hyper_params
        self.embeddings = Embedding(self.hp.vocab_size,self.hp.embed_units,mask_zero = True)

        self.bilstm1 = Bidirectional(LSTM(self.hp.num_units,return_sequences=True,dropout=self.hp.dropout))
        self.bilstm2 = Bidirectional(LSTM(self.hp.num_units,return_sequences=True,dropout=self.hp.dropout))

        self.mlp_layer = [Dense(self.hp.mlp_num_units,activation=self.hp.mlp_activation) for _ in range(self.hp.num_mlp_layer)]
        self.mlp_layer.append(Dense(self.hp.mlp_num_fan_out,activation=self.hp.mlp_activation))
        
        self.dropout=Dropout(rate=self.hp.dropout)

        #预测模型，采用hinge_loss,也可以考虑分类模型
        self.dense=Dense(1,activation='linear')
        

    def build(self,input_shape):
        if self.hp.bilinear:
            self.bilinear = self.add_weight(name = "bilinear weight",shape = (self.hp.embed_units,self.hp.embed_units),initializer = 'glorot_uniform',trainable = True)

        super(MVLSTM,self).build(input_shape)
        
    def call(self,inputs):
        """inputs:[left,right],left:[batch_size,T1],right:[batch_size,T2]"""
        
        #inputs作为列表输入
        left,right=inputs[0],inputs[1]
        
        left = self.embedding(left)
        right = self.embedding(right)
        
        left = self.bilstm1(left)#[batch_size,T1,embed_size]
        right = self.bilstm2(right)#[batch_size,T2,embed_size]
        
        match_matrix = Dot(axes=[2,2],normalize = False)([left,right])
        
        matching_signals=Reshape((-1,))(match_matrix)



import time

def get_local_time(is_print = True):
    '''
    返回当前本地时间字符串格式
    is_print = True:打印当前时间
    '''
    local_time = time.strftime("%Y-%m-%d %H:%M:%S",time.localtime())
    if is_print:
        print(local_time)
    return local_time





#from toolsbyerazhan import *#无效
#from toolsbyerazhan import timetools,gpu4tftools#有效

#可直接用timetools.py和gputools.py文件中的所有函数
#from .timetools import *
#from .gputools import *

from . import gputools,jsontools,layers,models,ostools,timetools
from . import pandastools,tensorflowtools,transformers

from .gputools import set_gpu_memory_tf
set_gpu_memory_tf()
from .quicktools import whether_to_transfer



import tensorflow as tf

#import tensorflow_addons as tfa
import tfa_crf
#tfa_crf文件是tfa.text中关于crf的一个文件
#可直接拿来用，不用安装整个tfa

from tensorflow.keras.layers import Layer,LSTM,Bidirectional,Embedding,Input,Dense,Dropout
from tensorflow.keras.initializers import glorot_uniform
import numpy as np

'''
参数总览：
inputs：维度是[batch_size,max_seq_length,embeded_size]，最简单方法是用ont-hot，
        但是怕单字非常多，可以考虑UNK，取出现次数较多的前N个单字，后面的用UNK代替
units：LSTM结构中h和c的大小
注意要加上每个batch中的长度，用于pad和mask
'''

class BiLSTMCRF(tf.keras.models.Model):
    '''vocab_size,单字个数，embed_size,词向量维度，units，隐藏层维度'''
    
    def __init__(self,vocab_size,embed_size,units,num_tags,*args,**kwargs):
        
        super(BiLSTMCRF,self).__init__()
        self.num_tags=num_tags
        self.embedding=Embedding(input_dim=vocab_size,output_dim=embed_size)
        self.bilstm=Bidirectional(LSTM(units,return_sequences=True),merge_mode='concat')
        #merge_mode的选择从维度角度是不影响输出结果的
        self.dense=Dense(num_tags)
        self.dropout=Dropout(0.5)
        
    def call(self,inputs):
        '''inputs维度：[batch_size,max_seq_length]'''
        inputs_length=tf.math.reduce_sum(tf.cast(tf.math.not_equal(inputs,0),dtype=tf.int32),axis=-1)
        #自动计算每个batch的seq_length，注意数据处理时pad=0
        
        x=self.embedding(inputs)
        x=self.bilstm(x)
        
        x=self.dropout(x)
        logits=self.dense(x)
        
        return logits,inputs_length
        
    def loss(self,logits,targets,inputs_length):
        targets=tf.cast(targets,dtype=tf.int32)

        #计算对数似然函数
        #log_likelihood,_=tfa.text.crf_log_likelihood(logits,targets,inputs_length,transition_params=self.transition_params)
        log_likelihood,_=tfa_crf.crf_log_likelihood(logits,targets,inputs_length,transition_params=self.transition_params)
        
        return log_likelihood,self.transition_params

    #定义转移矩阵transition_params
    def build(self,input_shape):
        shape=tf.TensorShape([self.num_tags,self.num_tags])
        self.transition_params=self.add_weight(name='transition_params',shape=shape,initializer=glorot_uniform,trainable=True)
        super(BiLSTMCRF,self).build(input_shape)

def test():
    batch_size=16
    vocab_size=100
    max_seq_length=8
    embed_size=32
    units=13
    num_tags=50
    
    inputs=tf.convert_to_tensor(np.random.randint(0,100,[batch_size,max_seq_length]),dtype=tf.int32)
    targets=tf.convert_to_tensor(np.random.randint(0,num_tags,[batch_size,max_seq_length]),dtype=tf.int32)
    my_model=BiLSTMCRF(vocab_size,embed_size,units,num_tags)
    
    logits,inputs_length=my_model(inputs)
    
    log_likelihood,transition_params=my_model.loss(logits,targets,inputs_length)

    #print(my_model.transition_params)
    return log_likelihood,transition_params
    
    
if __name__=="__main__":
    
    ll,tp=test()
    print(ll.shape)
    print(tp.shape)



def whether_to_transfer(x:float,y:float):
    
    v1 = int(x * 37700)
    v2 = int(y * 50500)
    diff = v1 - v2

    print("value of 601699:", v1)
    print("value of 601666:", v2)
    print("the difference is:", diff)
    
    if diff > 0:
        print("Congratulations! It is time to transfer.")
    else:
        print("It is not the right time to transfer, please be patient and wait!")




import tensorflow as tf

def test_shape(bsz = None, seq_len = 4, hidden_size = 5):
    tips = """
tips:
    params:
    bsz:None,ans is equal to static shape else dynamic shape
    seq_len:4
    hidden_size:6
    """
    if bsz == None:
        from tensorflow.keras import Input
        x = Input((seq_len,hidden_size),dtype = tf.float32)
    else:
        x = tf.range(bsz * seq_len * hidden_size)
        x = tf.reshape(x,[bsz,seq_len,hidden_size])
        x = tf.cast(x,tf.float32)

    static = x.shape.as_list()
    dynamic = tf.shape(x)
    ans = [dynamic[i] if s is None else s for i, s in enumerate(static)]
    print(tips)
    print(ans)
if __name__ == "__main__":
    test_shape()



#tensorflowtools专门用于理解tensorflow中的函数
#用__代替.例如tf.linalg.band_part用tf__linalg__band_part代替
#一个.py文件只有一个函数,文件名对于tf中从tf目录到具体函数名,
#它内部的函数命名规则是test_函数名,例如test_band_part

from .tf__keras__layers import test_Activation,test_Dot,test_Dropout,test_Reshape
from .tf__linalg__band_part import test_band_part
from .tf__linalg__matmul import test_matmul
from .tf__shape import test_shape



from setuptools import setup, find_packages

setup(
    name = "toolsbyerazhan",
    version = "0.1.13",

    keywords = ( "toolsbyerazhan"),
    description = "some useful tools",
    long_description = "some useful common tools",
    license = "MIT Licence",

    url = "https://github.com/erazhan/CommonTools",
    author = "erazhan",
    author_email = "erazhan@163.com",
    
    python_requires='>=3.7.1',#不需要太高的版本
    packages = find_packages(),
    include_package_data = True,
    platforms = "any",
    install_requires = ["tensorflow==2.2.0","pandas"]
)



import tensorflow as tf
import numpy as np

#from tensorflow.keras import Input
#from tensorflow.keras.models import Model
#from tensorflow.keras.layers import Layer,Embedding

#from xlnet_layers import EmbeddingRet,EmbeddingSim

#from xlnet_configuration import XLNetConfig 
from .xlnet_configuration import XLNetConfig
from .xlnet_utils import get_initializer,shape_list

#尽量保证,源码中有的都不删除,放在那就行,注意有一些是必须占位置的就赋值为None
'''
def gelu(x):
    return 0.5 * x * (1.0 + tf.math.erf(x / np.sqrt(2.0)))
'''
def gelu(x):
    """ Implementation of the gelu activation function.
        XLNet is using OpenAI GPT's gelu
        Also see https://arxiv.org/abs/1606.08415
    """
    cdf = 0.5 * (1.0 + tf.tanh((np.sqrt(2 / np.pi) * (x + 0.044715 * tf.pow(x, 3)))))
    return x * cdf

def swish(x):
    return x * tf.sigmoid(x)


ACT2FN = {
    "gelu": tf.keras.layers.Activation(gelu),
    "relu": tf.keras.activations.relu,
    "swish": tf.keras.layers.Activation(swish),
}

'''
#后续依次加上
custom_objects = {
    'gelu':gelu
    }
#也可以直接用赋值的方式
tf.keras.utils.get_custom_objects().update(custom_objects)

#最好都是把pad设置为0,可以 省去很多麻烦
#build_xlnet中mask_index!=0的情况暂时不考虑？
#这个不记得在哪里用了
'''
#{XLNetConfig,}注释掉的是原版

'''
#测试自定义的XLNetConfig
#config_class = XLNetConfig()
#print(config_class.vocab_size)
'''



class TFXNetRelativeAttention(tf.keras.layers.Layer):
    def __init__(self,config,**kwargs):
        super().__init__(**kwargs)
        if config.d_model % config.n_head != 0:
            raise ValueError(
                "The hidden size (%d) is not a multiple of the number of attention "
                "heads (%d)" % (config.d_model, config.n_head)
            )

        self.n_head = config.n_head
        self.d_head = config.d_head
        self.d_model = config.d_model
        self.scale = 1 / (config.d_head ** 0.5)#Attention之后softmax之前进行的scale
        self.initializer_range = config.initializer_range#各种训练参数的初始化方式

        self.layer_norm = tf.keras.layers.LayerNormalization(epsilon = config.layer_norm_eps, name = "layer_norm")
        self.dropout = tf.keras.layers.Dropout(config.dropout)
    
    def build(self, input_shape):
        initializer = get_initializer(self.initializer_range)#transformer:(d_model,d_model),xlnet:(d_model, n_head, d_head),应该是为了做head_mask
        self.q = self.add_weight(
            shape=(self.d_model, self.n_head, self.d_head), initializer=initializer, trainable=True, name="q"
        )
        self.k = self.add_weight(
            shape=(self.d_model, self.n_head, self.d_head), initializer=initializer, trainable=True, name="k"
        )
        self.v = self.add_weight(
            shape=(self.d_model, self.n_head, self.d_head), initializer=initializer, trainable=True, name="v"
        )
        self.o = self.add_weight(
            shape=(self.d_model, self.n_head, self.d_head), initializer=initializer, trainable=True, name="o"
        )
        self.r = self.add_weight(#用于相对位置向量
            shape=(self.d_model, self.n_head, self.d_head), initializer=initializer, trainable=True, name="r"
        )
        self.r_r_bias = self.add_weight(#相对位置向量偏置项
            shape=(self.n_head, self.d_head), initializer="zeros", trainable=True, name="r_r_bias"
        )
        self.r_s_bias = self.add_weight(#seg相对位置向量偏置项
            shape=(self.n_head, self.d_head), initializer="zeros", trainable=True, name="r_s_bias"
        )
        self.r_w_bias = self.add_weight(#词向量偏置项
            shape=(self.n_head, self.d_head), initializer="zeros", trainable=True, name="r_w_bias"
        )
        self.seg_embed = self.add_weight(#seg相对位置词向量
            shape=(2, self.n_head, self.d_head), initializer=initializer, trainable=True, name="seg_embed"
        )##和token_type_ids相关,只有2个embedding，在同一个seg中用第1个，不在同一个seg中(前后句)用第2个，注意这和Bert中的(绝对)位置向量也是不同的
        super().build(input_shape)

    def rel_shift(self, x, klen=-1):##用于调整bd(相对位置向量)的维度
        """perform relative shift to form the relative attention score."""
        x_size = shape_list(x)
        #只有这样转换才会和想要的结果一致,test_xlnet_function有测试函数
        x = tf.reshape(x, (x_size[1], x_size[0], x_size[2], x_size[3]))
        x = x[1:, ...]
        x = tf.reshape(x, (x_size[0], x_size[1] - 1, x_size[2], x_size[3]))
        x = x[:, 0:klen, :, :]
        
        return x

    def rel_attn_core(self, inputs, training=False):
        """Core relative positional attention operations."""
        #q_head:[qlen or num_predict,bsz,n_head,d_head],#qlen是context stream,num_predict是query stream
        #k_head_r:[klen(+1) or klen+qlen,bsz,n_head,d_head],
        #k(v)_head_h:[klen,bsz,n_head,d_head],
        #seg_mat:[qlen,klen,bsz,2]
        #attn_mask:[qlen,klen,bsz,1(d_model)]
        #head_mask = None
        #output_attentions= False
        q_head, k_head_h, v_head_h, k_head_r, seg_mat, attn_mask, head_mask, output_attentions = inputs

        # content based attention score
        ac = tf.einsum("ibnd,jbnd->ijbn", q_head + self.r_w_bias, k_head_h)##[qlen,klen,bsz,n_head]#如果不使用memory,klen=qlen

        # position based attention score
        bd = tf.einsum("ibnd,jbnd->ijbn", q_head + self.r_r_bias, k_head_r)##[qlen,klen+1 or klen+qlen,bsz,n_head]
        bd = self.rel_shift(bd, klen=shape_list(ac)[1])##bd的第二个维度上先去掉距离为klen的然后选klen-1->0,总的来说就是从klen,(klen-1),...,1,0,-1,...,-(qlen-1)中选(klen-1),...,1,0,变成了[qlen,klen,bsz,n_head]

        # segment based attention score
        if seg_mat is None:
            ef = 0
        else:##self.seg_embed就在build中定义了,s代表了2
            ef = tf.einsum("ibnd,snd->ibns", q_head + self.r_s_bias, self.seg_embed)#erazhan:这两步实际上是先用seg_mat与seg_embed作用生成位置词嵌入,然后再用q_head与位置词嵌入作用生成ef,这里的实现其实就是顺序不一样
            ef = tf.einsum("ijbs,ibns->ijbn", seg_mat, ef)##[qlen,klen,bsz,n_head]

        # merge attention scores and perform masking
        attn_score = (ac + bd + ef) * self.scale
        if attn_mask is not None:#erazhan:attn_mask维度就是[qlen,klen,bsz,1(d_model)]
            # attn_score = attn_score * (1 - attn_mask) - 1e30 * attn_mask
            attn_score = attn_score - 1e30 * attn_mask#dtype = tf.float32

        # attention probability
        attn_prob = tf.nn.softmax(attn_score, axis=1)

        attn_prob = self.dropout(attn_prob, training=training)

        # Mask heads if we want to##erazhan应该是用不上的,暂时不管
        if head_mask is not None:
            attn_prob = attn_prob * head_mask

        # attention output#erazhan:[qlen,klen,bsz,n_head]*[klen,bsz,n_head,d_head]->[qlen,bsz,n_head,d_head]
        attn_vec = tf.einsum("ijbn,jbnd->ibnd", attn_prob, v_head_h)
        
        if output_attentions is True:
            return attn_vec, attn_prob

        return attn_vec
    
    def post_attention(self, inputs, residual = True, training = False):##attention结束后,再经过一个o_dense层
        """Post-attention processing."""
        # post-attention projection (back to `d_model`)
        h, attn_vec = inputs

        attn_out = tf.einsum("ibnd,hnd->ibh", attn_vec, self.o)

        attn_out = self.dropout(attn_out, training=training)

        if residual:##实现残差连接,必须使用,h就是经过attention层之前的数据
            attn_out = attn_out + h
        output = self.layer_norm(attn_out)##加一层LN

        return output

    def call(self, inputs, training=False):
        (h, g, attn_mask_h, attn_mask_g, r, seg_mat, mems, target_mapping, head_mask, output_attentions) = inputs
        #h是词向量[qlen,bsz,d_model],g:[qlen,bsz,d_model],target_mapping:[num_predict,qlen?,bsz]
        #attn_mask_h就是non_tgt_mask,attn_mask_g就是attn_mask
        if g is not None:#h,g分别对应context stream,query stream,h是一定会有的,r就是pos_embed位置向量[klen(+1?单向) klen+qlen(双向),bsz,d_model]
            # Two-stream attention with relative positional encoding.
            # content based attention score
            if mems is not None and len(shape_list(mems)) > 1:##使用mems,如果两句话是上下句关系,那么会使用mems,否则不使用(训练语料会按0.5的概率选择上下句)
                cat = tf.concat([mems, h], axis=0)##mems:[mlen,bsz,d_model]
            else:
                cat = h##后续用cat

            # content-based key head
            k_head_h = tf.einsum("ibh,hnd->ibnd", cat, self.k)##[qlen or klen,batch_size,n_head,d_head]

            # content-based value head
            v_head_h = tf.einsum("ibh,hnd->ibnd", cat, self.v)##[qlen or klen,batch_size,n_head,d_head]

            # position-based key head
            k_head_r = tf.einsum("ibh,hnd->ibnd", r, self.r)##[klen(+1) or klen+qlen,batch_size,n_head,d_head]

            # h-stream
            # content-stream query head
            q_head_h = tf.einsum("ibh,hnd->ibnd", h, self.q)#[qlen,batch_size,n_head,d_head]#不考虑mems

            # core attention ops
            attn_vec_h = self.rel_attn_core(
                [q_head_h, k_head_h, v_head_h, k_head_r, seg_mat, attn_mask_h, head_mask, output_attentions],
                training=training,
            )

            if output_attentions is True:
                attn_vec_h, attn_prob_h = attn_vec_h

            # post processing
            output_h = self.post_attention([h, attn_vec_h], training=training)

            # g-stream
            # query-stream query head
            q_head_g = tf.einsum("ibh,hnd->ibnd", g, self.q)#[num_predict,bsz,n_head,d_head]它和q_head_h共用self.q,所有两个stream相互是有影响的

            # core attention ops
            if target_mapping is not None:#erazhan:只有query steam才有,这部分理解了,target_mapping在下面用了两次,1次将[num_predict,bsz,n_head,d_head]->[qlen,bsz,n_head,d_head],另1次效果相反
                q_head_g = tf.einsum("mbnd,mlb->lbnd", q_head_g, target_mapping)#每个能决定num_predict个token的正文token用num_predict个token的embedding表示求和，有点像Embedding层
                attn_vec_g = self.rel_attn_core(##只需计算num_predict个token的embedding表示就行了
                    [q_head_g, k_head_h, v_head_h, k_head_r, seg_mat, attn_mask_g, head_mask, output_attentions],
                    training=training,
                )

                if output_attentions is True:
                    attn_vec_g, attn_prob_g = attn_vec_g

                attn_vec_g = tf.einsum("lbnd,mlb->mbnd", attn_vec_g, target_mapping)##得到qlen个token的embedding后，再映射到num_predict维度(就是将seq_length个embedding求和)
            else:#没有target_mapping
                attn_vec_g = self.rel_attn_core(
                    [q_head_g, k_head_h, v_head_h, k_head_r, seg_mat, attn_mask_g, head_mask, output_attentions],
                    training=training,
                )

                if output_attentions is True:
                    attn_vec_g, attn_prob_g = attn_vec_g

            # post processing
            output_g = self.post_attention([g, attn_vec_g], training=training)

            if output_attentions is True:
                attn_prob = attn_prob_h, attn_prob_g

        else:#g is None也就是只有content stream,finetune阶段
            # Multi-head attention with relative positional encoding
            if mems is not None and len(shape_list(mems)) > 1:
                cat = tf.concat([mems, h], axis=0)
            else:
                cat = h

            # content heads
            q_head_h = tf.einsum("ibh,hnd->ibnd", h, self.q)
            k_head_h = tf.einsum("ibh,hnd->ibnd", cat, self.k)
            v_head_h = tf.einsum("ibh,hnd->ibnd", cat, self.v)

            # positional heads
            k_head_r = tf.einsum("ibh,hnd->ibnd", r, self.r)

            # core attention ops
            attn_vec = self.rel_attn_core(
                [q_head_h, k_head_h, v_head_h, k_head_r, seg_mat, attn_mask_h, head_mask, output_attentions],
                training=training,
            )

            if output_attentions is True:
                attn_vec, attn_prob = attn_vec

            # post processing
            output_h = self.post_attention([h, attn_vec], training=training)
            output_g = None

        outputs = (output_h, output_g)
        if output_attentions is True:
            outputs = outputs + (attn_prob,)
        return outputs
    

class TFXLNetFeedForward(tf.keras.layers.Layer):
    def __init__(self, config, **kwargs):
        super().__init__(**kwargs)
        self.layer_norm = tf.keras.layers.LayerNormalization(epsilon=config.layer_norm_eps, name="layer_norm")
        self.layer_1 = tf.keras.layers.Dense(
            config.d_inner, kernel_initializer=get_initializer(config.initializer_range), name="layer_1"
        )
        self.layer_2 = tf.keras.layers.Dense(
            config.d_model, kernel_initializer=get_initializer(config.initializer_range), name="layer_2"
        )
        self.dropout = tf.keras.layers.Dropout(config.dropout)
        if isinstance(config.ff_activation, str):
            self.activation_function = ACT2FN[config.ff_activation]
        else:
            self.activation_function = config.ff_activation

    def call(self, inp, training=False):
        output = inp
        output = self.layer_1(output)
        output = self.activation_function(output)
        output = self.dropout(output, training=training)
        output = self.layer_2(output)
        output = self.dropout(output, training=training)
        output = self.layer_norm(output + inp)
        return output

class TFXLNetLayer(tf.keras.layers.Layer):
    def __init__(self, config, **kwargs):
        super().__init__(**kwargs)
        self.rel_attn = TFXLNetRelativeAttention(config, name="rel_attn")
        self.ff = TFXLNetFeedForward(config, name="ff")
        self.dropout = tf.keras.layers.Dropout(config.dropout)

    def call(self, inputs, training=False):
        outputs = self.rel_attn(inputs, training=training)
        output_h, output_g = outputs[:2]

        if output_g is not None:
            output_g = self.ff(output_g, training=training)
        output_h = self.ff(output_h, training=training)
        
        outputs = (output_h, output_g)# + outputs[2:]  # Add again attentions if there are there # outputs[2:]暂时先不管
        return outputs

class TFXLNetLMHead(tf.keras.layers.Layer):
    #这个是预训练阶段预测时,用了共享的Embedding后,还需要加上一个偏置项
    def __init__(self,config,input_embeddings,**kwargs):
        super().__init__(**kwargs)
        self.vocab_size = config.vocab_size
        self.input_embeddings = input_embeddings

    def build(self,input_shape):
        self.bias = self.add_weight(shape = (self.vocab_size,),initializer = "zeros", trainable = True,name = "bias")
        super().build(input_shape)

    def call(self, hidden_states):
        #hidden_states:[bsz,qlen,d_model]->[bsz,qlen,vocab_size]->[bsz,qlen,vocab_size](add bias)
        hidden_states = self.input_embeddings(hidden_states,model="linear")
        hidden_states = hidden_states + self.bias
        return hidden_states

class TFXLNetMainLayer(tf.keras.layers.Layer):
    '''config参数在其它地方赋值'''
    #config_class = XLNetConfig
    def __init__(self,config,**kwargs):
        super().__init__(**kwargs)
        self.output_hidden_states = config.output_hidden_states
        self.oupput_attentions = config.output_attentions
        self.return_dict = config.return_dict#这是个啥

        self.mem_len = config.mem_len
        self.reuse_len = config.reuse_len
        self.d_model = config.d_model
        self.same_length = config.same_length#保留吧,在create_uni_mask中使用到
        self.attn_type = config.attn_type
        self.bi_data = config.bi_data#不考虑
        self.clamp_len = config.clamp_len
        self.n_layer = config.n_layer
        #self.use_bfloat16 = config.use_bfloat16#统统用tf.float32
        self.initializer_range = config.initializer_range

        self.word_embedding = TFSharedEmbeddings(
            config.vocab_size, config.d_model, initializer_range=config.initializer_range, name="word_embedding"
        )
        self.layer = [TFXLNetLayer(config, name="layer_._{}".format(i)) for i in range(config.n_layer)]
        self.dropout = tf.keras.layers.Dropout(config.dropout)
    def get_input_embeddings(self):
        return self.word_embedding

    def set_input_embeddings(self,value):
        self.word_embedding.weight = value
        self.word_embedding.vocab_size = value.shape[0]
    
    def build(self,input_shape):

        initializer = get_initializer(self.initializer_range)
        #在modeling_tf_utils中定义

        #这个用于构造初始query stream中的g,对应[num_predict,bsz,d_model],表示要预测token的词向量表示
        self.mask_emb = self.add_weight(shape = (1,1,self.d_model),initializer = initializer, trainable = True, name = "mask_emb")

    def create_uni_mask(self,qlen, mlen, dtype = tf.float32):
        '''单向模型时产生初始的attn_mask,此时不需要双向模型的perm_mask'''
        
        """
        Creates causal attention mask. Float mask where 1.0 indicates masked, 0.0 indicates not-masked.

        Args:
            qlen: TODO Lysandre didn't fill
            mlen: TODO Lysandre didn't fill

        ::

                  same_length=False:      same_length=True:
                  <mlen > <  qlen >       <mlen > <  qlen >
               ^ [0 0 0 0 0 1 1 1 1]     [0 0 0 0 0 1 1 1 1]
                 [0 0 0 0 0 0 1 1 1]     [1 0 0 0 0 0 1 1 1]
            qlen [0 0 0 0 0 0 0 1 1]     [1 1 0 0 0 0 0 1 1]
                 [0 0 0 0 0 0 0 0 1]     [1 1 1 0 0 0 0 0 1]
               v [0 0 0 0 0 0 0 0 0]     [1 1 1 1 0 0 0 0 0]

        """
        attn_mask = tf.ones([qlen, qlen], dtype=dtype)
        mask_u = tf.linalg.band_part(attn_mask, 0, -1)
        mask_dia = tf.linalg.band_part(attn_mask, 0, 0)
        attn_mask_pad = tf.zeros([qlen, mlen], dtype=dtype)
        ret = tf.concat([attn_mask_pad, mask_u - mask_dia], 1)
        if self.same_length:
            mask_l = tf.linalg.band_part(attn_mask, -1, 0)
            ret = tf.concat([ret[:, :qlen] + mask_l - mask_dia, ret[:, qlen:]], 1)
        return ret

    def cache_mem(self, curr_out, prev_mem):
        #关于mem的用法还不是很懂
        """cache hidden states into memory."""
        if self.reuse_len is not None and self.reuse_len > 0:
            curr_out = curr_out[: self.reuse_len]

        if prev_mem is None:
            new_mem = curr_out[-self.mem_len :]
        else:
            new_mem = tf.concat([prev_mem, curr_out], 0)[-self.mem_len :]

        return tf.stop_gradient(new_mem)

    @staticmethod
    def positional_embedding(pos_seq, inv_freq, bsz=None):
        sinusoid_inp = tf.einsum("i,d->id", pos_seq, inv_freq)
        pos_emb = tf.concat([tf.sin(sinusoid_inp), tf.cos(sinusoid_inp)], axis=-1)
        pos_emb = pos_emb[:, None, :]

        if bsz is not None:
            pos_emb = tf.tile(pos_emb, [1, bsz, 1])

        return pos_emb

    def relative_positional_encoding(self, qlen, klen, bsz=None, dtype = tf.float32):
        """create relative positional encoding."""
        freq_seq = tf.range(0, self.d_model, 2.0)
        freq_seq = tf.cast(freq_seq, dtype=dtype)

        inv_freq = 1 / (10000 ** (freq_seq / self.d_model))
        assert self.att_type == "bi" or self.att_type == "uni", "att_type must be 'bi' or 'uni'"
        if self.attn_type == "bi":
            # beg, end = klen - 1, -qlen
            beg, end = klen, -qlen
        else:#self.attn_type == "uni"
            # beg, end = klen - 1, -1
            beg, end = klen, -1

        fwd_pos_seq = tf.range(beg, end, -1.0)
        fwd_pos_seq = tf.cast(fwd_pos_seq, dtype = dtype)
        if self.clamp_len > 0:
            fwd_pos_seq = tf.clip_by_value(fwd_pos_seq, -self.clamp_len, self.clamp_len)
        pos_emb = self.positional_embedding(fwd_pos_seq, inv_freq, bsz)#bsz仅仅用处只有一个

        return pos_emb

    def call(self,
             inputs,
             attention_mask = None,
             mems = None,
             perm_mask = None,
             target_mapping = None,
             token_type_ids = None,
             input_mask = None,
             head_mask = None,#这个暂不考虑,
             inputs_embeds = None,
             use_cache = True,
             output_attentions = None,
             output_hidden_states = None,
             #return_dict = None,#这个是新版本的,暂不考虑
             training = False
             ):
        #判断inputs格式,至少有input_ids,这里暂不考虑dict,BatchEncoding格式(源码中的第2种)
        if isinstance(inputs, (tuple,list)):
            #输入严格按照顺序来
            input_ids = inputs[0]#[bsz,qlen]
            attention_mask = inputs[1] if len(inputs) > 1 else attention_mask#[bsz,qlen,qlen],0代表mask
            mems = inputs[2] if len(inputs) > 2 else mems#列表,个数是attention的层数,后续补充使用说明
            perm_mask = inputs[3] if len(inputs) > 3 else perm_mask#[bsz,qlen,qlen(klen?)]
            target_mapping = inputs[4] if len(inputs) > 4 else target_mapping#维度[bsz,num_predict,qlen(klen)]
            token_type_ids = inputs[5] if len(inputs) > 5 else token_type_ids#[bsz,qlen]
            input_mask = inputs[6] if len(inputs) > 6 else input_mask#[bsz,qlen],和attention_mask相反,1代表mask
            head_mask = inputs[7] if len(inputs) > 7 else head_mask#对不同的head进行mask,暂不考虑,直接填个None就行
            inputs_embeds = inputs[8] if len(inputs) > 8 else inputs_embeds#[bsz,qlen,d_model(hidden_size)]就是word_embedding(input_ids)
            use_cache = inputs[9] if len(inputs) > 9 else use_cache#是否使用memory机制,默认True
            output_attentions = inputs[10] if len(inputs) > 10 else output_attentions#是否输出每个attention层后的qk(还没到qkv),暂不考虑
            output_hidden_states = inputs[11] if len(inputs) > 11 else output_hidden_states##输出每一层的结果(后续再完善理解),暂不考虑
            assert len(inputs) <= 12, "Too many inputs."
        else:
            input_ids = inputs

        #放这里,先不考虑
        #output_attentions = output_attentions if output_attentions is not None else self.output_attentions
        #output_hidden_states = output_hidden_states if output_hidden_states is not None else self.output_hidden_states
        
        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("You cannot specify both input_ids and inputs_embeds at the same time")
        elif input_ids is not None:#[bsz,qlen]->[qlen,bsz]
            input_ids = tf.transpose(input_ids, perm=(1, 0))
            qlen, bsz = shape_list(input_ids)[:2]
        elif inputs_embeds is not None:#[bsz,qlen,d_model]->[qlen,bsz,d_model]
            inputs_embeds = tf.transpose(inputs_embeds, perm=(1, 0, 2))
            qlen, bsz = shape_list(inputs_embeds)[:2]
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds")

        token_type_ids = tf.transpose(token_type_ids, perm=(1, 0)) if token_type_ids is not None else None#[bsz,qlen]->[qlen,bsz]
        input_mask = tf.transpose(input_mask, perm=(1, 0)) if input_mask is not None else None#[bsz,qlen]->[qlen,bsz]
        attention_mask = tf.transpose(attention_mask, perm=(1, 0)) if attention_mask is not None else None#[bsz,qlen]->[qlen,bsz]
        perm_mask = tf.transpose(perm_mask, perm=(1, 2, 0)) if perm_mask is not None else None#[bsz,qlen,qlen]->[qlen,qlen,bsz]#第2个确定是qlen,后续计算好data_mask后加上mlen
        target_mapping = tf.transpose(target_mapping, perm=(1, 2, 0)) if target_mapping is not None else None#[bsz,num_predict,qlen]->[num_predict,qlen,bsz]

        mlen = shape_list(mems[0])[0] if mems is not None and mems[0] is not None else 0#mlen和memory有关
        klen = mlen + qlen

        # Attention mask
        # causal attention mask
        if self.attn_type == "uni":#选择单向模型,此处得到的atten_mask其实和Transformer的decoder的mask一样，注意该符号和attention_mask的区别
            attn_mask = self.create_uni_mask(qlen, mlen)
            attn_mask = attn_mask[:, :, None, None]#[qlen,klen]->[qlen,klen,1(bsz),1(d_model or d_head?)]
        elif self.attn_type == "bi":#选择双向模型,此时不用attn_mask,而是用perm_mask,二者只需一个
            attn_mask = None
        else:
            raise ValueError("Unsupported attention type: {}".format(self.attn_type))

        # data mask: input mask & perm mask
        assert input_mask is None or attention_mask is None, (#input_mask和attention_mask二选一
            "You can only use one of input_mask (uses 1 for padding) "
            "or attention_mask (uses 0 for padding, added for compatbility with BERT). Please choose one."
        )
        if input_mask is None and attention_mask is not None:
            input_mask = 1.0 - tf.cast(attention_mask, dtype = tf.float32)#input_mask中1是被mask,后续只用input_mask
        if input_mask is not None and perm_mask is not None:#使用input_mask和perm_mask,Tensor加个None效果是增加1个第0维,原来的维度向后移一维
            data_mask = input_mask[None] + perm_mask#维度为[qlen,qlen,bsz],data_mask可能出现2,最后转为1.0
        elif input_mask is not None and perm_mask is None:#没有perm_mask
            data_mask = input_mask[None]
        elif input_mask is None and perm_mask is not None:#没有input_mask
            data_mask = perm_mask
        else:
            data_mask = None#两个mask都没有
     
        if data_mask is not None:
            # all mems can be attended to
            if mlen > 0:#加上memory,对应的mask全为0,也就是不mask
                mems_mask = tf.zeros([shape_list(data_mask)[0], mlen, bsz], dtype = tf.float32)#[qlen,mlen,bsz]
                data_mask = tf.concat([mems_mask, data_mask], axis=1)#[qlen,klen,bsz]
            if attn_mask is None:#双向
                attn_mask = data_mask[:, :, :, None]
            else:#单向,此时data_mask中的perm_mask应该不用给定
                attn_mask += data_mask[:, :, :, None]#data_mask和attn_mask相加,保留attn_mask

        if attn_mask is not None:
            attn_mask = tf.cast(attn_mask > 0, dtype = tf.float32)

        if attn_mask is not None:##将attn_mask->non_tgt_mask,就是将对角线上元素设置为0,被perm_mask掉的token在计算权重时是需要考虑自身token对自己产生的影响的,比如将最后两个词作为预测token
            non_tgt_mask = -tf.eye(qlen, dtype = tf.float32)##对角线上的全部设置为0，消除当前token从1变成0，说实话还不是很懂为什么要这么做,感觉是多余的?
            if mlen > 0:
                non_tgt_mask = tf.concat([tf.zeros([qlen, mlen], dtype = tf.float32), non_tgt_mask], axis=-1)
            non_tgt_mask = tf.cast((attn_mask + non_tgt_mask[:, :, None, None]) > 0, dtype = tf.float32)
        else:
            non_tgt_mask = None
        #后续使用attn_mask和non_tgt_mask,在双向模式时,attn_mask用于query stream,non_tgt_mask用于contest stream
        
        # Word embeddings and prepare h & g hidden states
        if inputs_embeds is not None:
            word_emb_k = inputs_embeds#[qlen,bsz,d_model]
        else:
            word_emb_k = self.word_embedding(input_ids)
        output_h = self.dropout(word_emb_k, training = training)#[qlen,bsz,d_model]#这个就是k_head_h
        if target_mapping is not None:#erazhan:[1,1,d_model] -> [num_predict,bsz,d_model],在预训练阶段我们只需要得到num_predict个token的最终词向量表示进行计算损失函数
            word_emb_q = tf.tile(self.mask_emb, [shape_list(target_mapping)[0], bsz, 1])
            #初始化用每个bsz每个token的表示用同一个向量表示    
            output_g = self.dropout(word_emb_q, training = training)
        else:
            output_g = None

        # Segment embedding
        if token_type_ids is not None:#token_type_ids:[qlen,bsz]
            # Convert `token_type_ids` to one-hot `seg_mat`
            if mlen > 0:
                mem_pad = tf.zeros([mlen, bsz], dtype=tf.int32)
                cat_ids = tf.concat([mem_pad, token_type_ids], 0)
            else:
                cat_ids = token_type_ids

            # `1` indicates not in the same segment [qlen x klen x bsz]##erazhan:[qlen,1,bsz] [1,qlen or klen,bsz]->[qlen,qlen or klen,bsz]
            seg_mat = tf.cast(tf.logical_not(tf.equal(token_type_ids[:, None], cat_ids[None, :])), tf.int32)#参考test_xlnet_functions中的test_seg_mat
            seg_mat = tf.one_hot(seg_mat, 2, dtype=dtype_float)##最后还会与2维的segment矩阵相乘，后续再看怎么用[qlen,klen,bsz,2]
        else:##seg_mat也是考虑相对位置的(也就是对于1个token对另一个token),在同一个seg中用相同的seg_embed,不同的seg则用另外一个(在bert中我理解的seg embed是在前后seg分别对应用一个词嵌入,没有考虑相对位置)
            seg_mat = None

        # Positional encoding
        pos_emb = self.relative_positional_encoding(qlen, klen, bsz = bsz, dtype = tf.float32)
        pos_emb = self.dropout(pos_emb, training = training)

        #head_mask暂时不管

        new_mems = ()#每层一个,mems[i]:[bsz,mlen,d_model](具体再看后续)
        if mems is None:
            mems = [None] * len(self.layer)

        attentions = []#[qlen,klen,bsz,d_model]?
        hidden_states = []#[qlen,bsz,d_model]?

        for i, layer_module in enumerate(self.layer):
            # cache new mems
            if self.mem_len is not None and self.mem_len > 0 and use_cache is True:
                new_mems = new_mems + (self.cache_mem(output_h, mems[i]),)
            #这个暂时先不管
            #if cast_bool_to_primitive(output_hidden_states) is True:
            #    hidden_states.append((output_h, output_g) if output_g is not None else output_h)

            outputs = layer_module(#self.layer TFXLNetRelativeAttention
                [
                    output_h,
                    output_g,
                    non_tgt_mask,
                    attn_mask,
                    pos_emb,
                    seg_mat,
                    mems[i],
                    target_mapping,
                    head_mask[i],
                    output_attentions,
                ],
                training=training,
            )
            output_h, output_g = outputs[:2]
            #暂不考虑
            #if cast_bool_to_primitive(output_attentions) is True:
            #    attentions.append(outputs[2])

        # Add last hidden state
        #if cast_bool_to_primitive(output_hidden_states) is True:
        #    hidden_states.append((output_h, output_g) if output_g is not None else output_h)

        #预训练阶段或者语言模型生成token用query stream的结果预测最后两个token,其它情况直接用最后的[bsz,qlen,d_model]作为token的词向量表示参与下游任务(和Bert类似,但是具体细节还要研究下)
        output = self.dropout(output_g if output_g is not None else output_h, training=training)

        # Prepare outputs, we transpose back here to shape [bsz, len, hidden_dim] (cf. beginning of forward() method)
        outputs = (tf.transpose(output, perm=(1, 0, 2)),)

        if self.mem_len is not None and self.mem_len > 0 and use_cache is True:
            outputs = outputs + (new_mems,)
        """
        if cast_bool_to_primitive(output_hidden_states) is True:
            if output_g is not None:
                hidden_states = tuple(tf.transpose(h, perm=(1, 0, 2)) for hs in hidden_states for h in hs)
            else:
                hidden_states = tuple(tf.transpose(hs, perm=(1, 0, 2)) for hs in hidden_states)
            outputs = outputs + (hidden_states,)
        if cast_bool_to_primitive(output_attentions) is True:
            attentions = tuple(tf.transpose(t, perm=(2, 3, 0, 1)) for t in attentions)
            outputs = outputs + (attentions,)
        """
        
        return outputs  # outputs, (new_mems), (hidden_states), (attentions)   

#PreTrainedModel
class TFXLNetModel(tf.keras.Model):
    def __init__(self,config,*inputs,**kwargs):
        super().__init__(*inputs,**kwargs)
        self.transformer = TFXLNetMainLayer(config, name = "transformer")

    def call(self,inputs,**kwargs):
        outputs = self.transformer(inputs,**kwargs)
        return outputs
        
class TFXNetLMHeadModel(tf.keras.Model):
    def __init__(self,config,*inputs,**kwargs):
        self.transformer = TFXLNetMainLayer(*config, name = "transformer")
        self.lm_loss = TFXLNetLMHead(config,self.transformer.word_embedding,name = "lm_loss")
    def get_output_embeddings(self):
        return self.lm_loss.input_embeddings

    def call(
        self,
        inputs,
        attention_mask = None,
        mems = None,
        perm_mask = None,
        target_mapping = None,
        token_type_ids = None,
        input_mask = None,
        head_mask = None,
        inputs_embeds = None,
        use_cache = True,
        output_attentions = None,
        output_hidden_states = None,
        labels = None,
        training = False,
        ):
        if isinstance(inputs, (tuple, list)):
            labels = inputs[12] if len(inputs) > 12 else labels
            if len(inputs) > 12:
                inputs = inputs[:12]
        elif isinstance(inputs, (dict, BatchEncoding)):
            labels = inputs.pop("labels", labels)

        transformer_outputs = self.transformer(
            inputs,
            attention_mask=None,
            mems=None,
            perm_mask=None,
            target_mapping=None,
            token_type_ids=None,
            input_mask=None,
            head_mask=None,
            inputs_embeds=None,
            use_cache=True,
            output_attentions=None,
            output_hidden_states=None,
            training=training,
        )
        hidden_state = transformer_outputs[0]
        logits = self.lm_loss(hidden_state, training=training)

        outputs = (logits,) + transformer_outputs[1:]  # Keep mems, hidden states, attentions if there are in it

        if labels is not None:
            # shift labels to the left and cut last logit token
            logits = logits[:, :-1]
            labels = labels[:, 1:]
            loss = self.compute_loss(labels, logits)
            outputs = (loss,) + outputs

        return outputs  # return logits, (mems), (hidden states), (attentions)

if __name__ == "__main__":
    import toolsbyerazhan as tbe
    tbe.set_gpu_memory_tf()





        
